/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2009 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 *
 *******************************************************************************
 *
 * $Id: ClassMap.java,v d60681e7fc1d 2009/09/17 23:59:23 fabrizio $
 *
 ******************************************************************************/
package it.tidalwave.metadata.spi;

import java.util.Comparator;
import java.util.Map;
import java.util.TreeMap;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;

/*******************************************************************************
 *
 * A {@link Map} whose key is a {@link Class}. The {@link #get(Object)} method returns 
 * a value even when there is not an exact match for the key, but a subclass of
 * the key exists in the map. Furthermore, if multiple classes in a hierarchy
 * are present and an exact match is not found, <code>get()</code> is guaranteed
 * to match the most specialized class.
 * 
 * For instance, consider to have classes <code>A</code>, <code>B</code> and
 * <code>C</code>, such as <code>B extends A</code> and <code>C extends B</code>;
 * and suppose to have:
 * 
 * <pre>
 * ClassMap&lt;String&gt; map = new ClassMap&lt;String&gt;();
 * map.put(A.class, "one");
 * map.put(B.class, "two");
 * </pre>
 * 
 * Then, the following holds true:
 * 
 * <pre>
 * String s1 = map.get(A.class); // returns "one"
 * String s2 = map.get(B.class); // returns "two"
 * String s3 = map.get(C.class); // returns "two"
 * </pre>
 * 
 * <code>ClassMap</code> is useful for registering services bound to a specific
 * class and generically to all of its subclasses (this e.g. happens when you use
 * a bytecode enhancer such as CGLIB that produces objects which are instances
 * of synthetic subclasses, like <code>C</code> in the example); and where it's 
 * possible that a specific subclass is bound to a specific service (like 
 * <code>B</code> in the example).
 *
 * @hidden
 *
 * @author  Fabrizio Giudici
 * @version $Id: ClassMap.java,v d60681e7fc1d 2009/09/17 23:59:23 fabrizio $
 * 
 ******************************************************************************/
public class ClassMap<T> extends TreeMap<Class<?>, T>
  {
    private static final long serialVersionUID = -8094417265456628888L;
    
    public ClassMap()
      {
        // This Comparator makes sure that entries are sorted by revered 
        // hierarchy depth, that is subclasses come first.
        super(new Comparator<Class<?>>() 
          {
            public int compare (@Nonnull final Class<?> c1, @Nonnull final Class<?> c2) 
              {
                final int sgn = (int)Math.signum(depth(c2) - depth(c1));
                
                if (sgn == 0) // be deterministic in any case
                  {
                    return c1.getName().compareTo(c2.getName());  
                  }
                
                return sgn; 
              }
            
            private int depth (@Nonnull final Class<?> clazz)
              {
                final Class<?> superClass = clazz.getSuperclass();
                return (superClass == null) ? 0 : depth(superClass) + 1;
              }
        }); 
      }

    /***************************************************************************
     * 
     * Returns the object bound to the given key.
     * 
     * The search is performed by just looking at all the entries in the map;
     * since they are sorted in reversal hyerarchy depth it's ok to return the
     * first match (any further match would refer to superclasses).
     *
     **************************************************************************/
    @Override
    @CheckForNull 
    public T get (@Nonnull final Object key) 
      {
        for (final Map.Entry<Class<?>, T> entry : entrySet())
          {
            if (entry.getKey().isAssignableFrom((Class)key))
              {
                return entry.getValue();  
              }
          }
        
        return null;
      }
  }
