/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2009 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 *
 *******************************************************************************
 *
 * $Id: MetadataItemEnhancer.java,v d60681e7fc1d 2009/09/17 23:59:23 fabrizio $
 *
 ******************************************************************************/
package it.tidalwave.metadata.spi;

import net.sf.cglib.proxy.Callback;
import it.tidalwave.beans.AbstractEnhancer;
import it.tidalwave.beans.JavaBean;
import it.tidalwave.metadata.spi.impl.MetadataItemAspect;
import java.beans.PropertyChangeListener;
import javax.annotation.Nonnull;

/*******************************************************************************
 * 
 * This factory provides enhanced metadata items that support the JavaBean
 * specification and implement the {@link MetadataItem} interface. The requisites
 * for the source classes are:
 * 
 * <ol>
 * <li>to provide properties by means of getter/setters;</li>
 * <li>to provide an "*Available" property for each property (e.g.
 *     <code>isFooAvailable()</code> for <code>getFoo()</code>);</li>
 * </ol>
 * 
 * <pre>
 * public class ExampleMetadataItem
 *   {
 *     private String name;
 *     private boolean nameAvailable;
 *     
 *     public void setName (final String name)
 *       {
 *         this.name = name;    
 *       }
 *     
 *     public String getName()
 *       {
 *         return name;    
 *       }
 *     
 *     public void setNameAvailable (final boolean nameAvailable)
 *       {
 *         this.nameAvailable = nameAvailable;    
 *       }
 *     
 *     public boolean isNameAvailable()
 *       {
 *         return nameAvailable;    
 *       }
 *   }
 * </pre>
 * 
 * Now, calling:
 * 
 * <pre>
 * ExampleMetadataItem item = MetadataItemEnhancer.createEnhancedItem(new ExampleMetadataItem()));
 * </pre>
 * 
 * an enhanced item is returned. To use the enhanced features such as the
 * support for {@link PropertyChangeListener}s, you have to use the reflection API
 * or to explicitly cast as in the example below:
 * 
 * <pre>
 * ((JavaBean)item).addPropertyChangeListener(...);
 * </pre>
 * 
 * In addition to the JavaBean specifications, the enhanced items also supports 
 * these semantics:
 * 
 * <ol>
 * <li>setting a property to a non-<code>null</code> value will force the related 
 *     <code>***Available()</code> property to <code>true</code>;</li>
 * <li>setting an <code>***Available()</code> property to <code>false</code> will 
 *     force the related property to <code>null</code>.
 * </ol>
 * 
 * @hidden
 *
 * @author  Fabrizio Giudici
 * @version $Id: MetadataItemEnhancer.java,v d60681e7fc1d 2009/09/17 23:59:23 fabrizio $
 *
 ******************************************************************************/
public class MetadataItemEnhancer extends AbstractEnhancer<JavaBean>
  {
    private final static MetadataItemEnhancer instance = new MetadataItemEnhancer();
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private MetadataItemEnhancer()
      {
        super(JavaBean.class);
      }
    
    /***************************************************************************
     *
     * Creates an enhanced item.
     * 
     * @param   bean  the original item
     * @return        the enhanced item
     * 
     **************************************************************************/
    @Nonnull 
    public static <T> T createEnhancedItem (@Nonnull final T bean, @Nonnull final Object ... arguments)
      {
        return (T)instance.createEnhancedBean(bean, arguments);
      }
    
    /***************************************************************************
     *
     * {@inheritDoc} This method specifies {@link MetadataItemAspect} as the
     * aspect implementation to give to enhanced beans.
     * 
     **************************************************************************/
    @Override
    @Nonnull 
    protected Callback createInterceptor (@Nonnull final Object bean, @Nonnull final Object ... arguments)
      {
        return new MetadataItemAspect(bean, this, arguments);
      }
  }
         